# -*- coding: utf-8 -*-
"""forms.py: appsearch forms"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import operator

from django import forms
from django.forms import ValidationError
from django.forms.formsets import BaseFormSet
from django.db.models.fields import BLANK_CHOICE_DASH
from django.contrib.contenttypes.models import ContentType

import dateutil.parser


class ModelSelectionForm(forms.Form):
    """
    Default Model selection form.

    Builds choices based on the registered configurations in the given ``registry``.  The form uses
    content type ids for values instead of model names or module paths, etc.

    """

    model = forms.ChoiceField(label="Search For")

    def __init__(self, registry, user, permission, *args, **kwargs):
        super(ModelSelectionForm, self).__init__(*args, **kwargs)

        self.registry = registry
        self.configurations = registry.get_configurations(user=user, permission=permission)

        self.fields['model'].choices = BLANK_CHOICE_DASH + \
            [(c._content_type.id, c.verbose_name) for c in self.configurations]

    def clean_model(self):
        """ Cleans the content type id into the model it represents. """
        model = self.cleaned_data['model']
        try:
            model = ContentType.objects.get(id=model).model_class()
        except ContentType.DoesNotExist as e:
            raise ValidationError("Invalid choice - {}".format(e))
        if model not in self.registry:
            raise ValidationError("Invalid choice")
        return model

    def get_selected_configuration(self):
        """
        Given that the form has passed validation, returns the configuration for the selected model.

        """
        model_value = self.cleaned_data['model']
        return self.registry[model_value]

    def get_selected_model(self):
        """ Given that the form has passed validation, returns the selected model. """
        return self.get_selected_configuration().model


class ConstraintForm(forms.Form):
    """
    Using an additional constructor parameter ``configuration``, an instance of
    ``registry.ModelSearch``, the form dynamically populates the choices for the ``field`` field.

    """

    type = forms.ChoiceField(label="Constraint", choices=[
        ('and', 'AND'),
        ('or', 'OR'),
    ])

    # Dynamically populated list of fields available for filtering
    field = forms.ChoiceField(label="Filter by", choices=[])

    # Dynamically populated list of valid operators for the chosen ``field``
    operator = forms.ChoiceField(label="Constraint type", choices=[])

    term = forms.CharField(label="Search term", required=False)
    end_term = forms.CharField(label="End term", required=False)

    def __init__(self, configuration, *args, **kwargs):
        """
        Receives the configuration for the model to represent (potentially ``None`` if the original
        ``ModelSelectionForm`` is blank or invalid).

        """

        super(ConstraintForm, self).__init__(*args, **kwargs)

        self.configuration = configuration
        if configuration:
            self.fields['field'].choices = configuration.get_searchable_field_choices()

    def _clean_fields(self):
        """
        Workaround for displaying stripped info in the ``operator`` select, and running validation
        against reversed (name, value) choices.

        Because the frontend strips away the queryset language from the operator choices (leaving
        behind doubled 2-tuples of UI text such as ("= equal", "= equal")), the default field-level
        validation of the ``operator`` field will fail by default, since "iexact" is valid but
        "= equal" is not.  To solve the problem, this method sets the choices appropriately in
        reverse: ("= equal", "iexact").  When field validation inspects the value, everything will
        check out.  During that time, the ``clean_operator()`` method will be called to further
        process the ``operator`` value back to "iexact" (the reversed, faked UI text).

        Once all of the trickery is done, the choices are mapped once more to ("= equal", "= equal")
        so that when the template is rendered with this form, the choices are as they were before.

        """

        field_field = self.fields['field']
        field_hash = field_field.widget.value_from_datadict(self.data, {}, self.add_prefix('field'))
        operators = self.configuration.get_operator_choices(hash=field_hash)
        self.fields['operator'].choices = map(list, map(reversed, operators))

        # Allow default validation to occur, including the "clean_operator" method below.
        super(ConstraintForm, self)._clean_fields()

        # Set the operators back to the flat choices of frontend-only values
        self.fields['operator'].choices = map(lambda o: (o[1], o[1]), operators)

    def clean_type(self):
        """ Convert type into an ``operator.and_`` or ``operator.or_`` reference. """

        type = self.cleaned_data['type']

        if type == "and":
            type = operator.and_
        elif type == "or":
            type = operator.or_

        return type

    def clean_field(self):
        """ Convert ``field`` hash into ORM path tuple. """
        return self.configuration.reverse_field_hash(self.cleaned_data['field'])

    def clean_operator(self):
        """ Convert operator into ORM query type such as "icontains" """

        # The default cleaned_value will be the text from the UI, so the choices (which have been
        # reversed for default validation to accept the text as a valid choice) are converted to a
        # dictionary for accessing the intended ORM queryset language value.

        # Input: "= equal"
        # Output: "iexact"

        operator = self.cleaned_data['operator']
        operator = dict(self.fields['operator'].choices)[operator]

        return operator

    def clean_term(self):
        """ Normalizes the ``term`` field to what makes sense for the operator. """

        if 'field' not in self.cleaned_data or 'operator' not in self.cleaned_data:
            return self.cleaned_data['term']

        classification = self.configuration.get_field_classification(self.cleaned_data['field'])
        operator = self.cleaned_data['operator']
        term = self.cleaned_data['term'].strip()
        field_type = self.configuration.field_types[self.cleaned_data['field']]

        if field_type.choices:
            # The field's database values aren't the display values, but the display values are
            # what the user will search for.  e.g., choices=[(1, 'Bob'), (2, 'Mary')]
            choices = {k.lower(): v for v, k in field_type.choices}
            if term.lower() in choices:
                term = choices[term.lower()]
        else:
            # Numbers and strings don't need processing, but the other types should be inspected.
            if classification == "date":
                try:
                    term = dateutil.parser.parse(term)
                except (TypeError, ValueError):
                    raise ValidationError("Unable to parse a date from '{}'".format(term))
            elif classification == "boolean":
                if term.lower() in ("true", "yes"):
                    term = True
                elif term.lower() in ("false", "no"):
                    term = False
                else:
                    raise ValidationError("Boolean value must be either true/false or yes/no.")
            elif classification == "number":
                try:
                    float(term)
                except:
                    raise ValidationError("Value must be numeric.")

        if operator not in ('isnull', '!isnull') and term in [None, '']:
            raise ValidationError("This field is required.")

        return term

    def clean_end_term(self):
        """
        Normalize the ``end_term`` field for range queries.

        The ``end_term`` will merge itself into ``term`` if appropriate, making the cleaned value
        of ``term`` a list of the two values suitable for direct use in a queryset "range" lookup.
        Consequently, ``end_term`` will always clean to the empty string.

        """

        if 'field' not in self.cleaned_data or 'operator' not in self.cleaned_data \
                or 'term' not in self.cleaned_data:
            return self.cleaned_data['end_term']

        classification = self.configuration.get_field_classification(self.cleaned_data['field'])
        operator = self.cleaned_data['operator']
        begin_term = self.cleaned_data['term']
        term = self.cleaned_data['end_term']

        if operator == "range":
            if classification == "date":
                term = dateutil.parser.parse(term)
            elif classification == "number":
                term = int(term)
            else:
                raise ValidationError("Unknown range type %r." % classification)
            self.cleaned_data['term'] = [begin_term, term]

        return ""


class ConstraintFormset(BaseFormSet):
    """ Removes the first ``ConstraintForm``'s ``type`` field. """

    def __init__(self, configuration, *args, **kwargs):
        """ Stores the configuration until the forms are constructed. """
        self.configuration = configuration
        super(ConstraintFormset, self).__init__(*args, **kwargs)

    def _construct_form(self, i, **kwargs):
        """ Sends the specified model configuration to the form. """
        return super(ConstraintFormset, self)._construct_form(i, configuration=self.configuration,
                **kwargs)



# -*- coding: utf-8 -*-
"""views.py: ORM Utils"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from django.views.generic import TemplateView

from appsearch.utils import Searcher


class SearchMixin(object):
    searcher_class = Searcher

    context_object_name = 'search'
    form_template_name = "appsearch/default_form.html"
    search_form_template_name = "appsearch/search_form.html"
    results_list_template_name = "appsearch/results_list.html"

    # Callbacks, unprovided by default
    get_display_fields = None
    build_queryset = None
    process_results = None

    def get_context_data(self, **kwargs):
        context = super(SearchMixin, self).get_context_data(**kwargs)

        object_name = self.get_context_object_name()
        searcher = self.get_searcher()

        if searcher.ready:
            searcher._perform_search()

        context[object_name] = searcher
        return context

    def get_searcher_class(self):
        """ Returns the view's ``searcher_class`` attribute. """
        return self.searcher_class

    def get_searcher(self):
        """ Builds and returns a ``Searcher`` instance for this search context. """
        return self.get_searcher_class()(self.request, **self.get_searcher_kwargs())

    def get_searcher_kwargs(self):
        """ Returns the dictionary of kwargs sent to the ``Searcher`` constructor. """

        return {
            'form_template_name': self.get_form_template_name(),
            'search_form_template_name': self.get_search_form_template_name(),
            'results_list_template_name': self.get_results_list_template_name(),
            'context_object_name': self.get_context_object_name(),

            # Callbacks
            'display_fields_callback': self.get_display_fields,
            'build_queryset_callback': self.build_queryset,
            'process_results_callback': self.process_results,
        }

    def get_context_object_name(self):
        return self.context_object_name

    def get_form_template_name(self):
        return self.form_template_name

    def get_search_form_template_name(self):
        return self.search_form_template_name

    def get_results_list_template_name(self):
        return self.results_list_template_name

class BaseSearchView(SearchMixin, TemplateView):
    pass





# This code is heavily based on contrib.admin's autodiscover mechanism.

import logging
import sys

__name__ = 'appsearch'
__author__ = 'Pivotal Energy Solutions'
__version_info__ = (1, 1, "0-rc0")
__version__ = '.'.join(map(str, __version_info__))
__date__ = '2014/07/22 4:47:00 PM'
__credits__ = ['Tim Valenta', "Steven Klass"]
__license__ = 'See the file LICENSE.txt for licensing information.'

log = logging.getLogger(__name__)

def autodiscover():
    """
    Auto-discover INSTALLED_APPS search.py modules and fail silently when
    not present. This forces an import on them to register any appsearch bits
    they may want.
    """

    import copy
    from importlib import import_module

    from django.conf import settings
    from django.utils.module_loading import module_has_submodule
    from appsearch.registry import search


    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's appsearch module.
        try:
            before_import_registry = copy.copy(search._registry)
            import_module('%s.search' % app)
        except:
            # Reset the model registry to the state before the last import as
            # this import will have to reoccur on the next request and this
            # could raise NotRegistered and AlreadyRegistered exceptions
            # (see #8245).
            search._registry = before_import_registry

            # Decide whether to bubble up this error. If the app just
            # doesn't have an admin module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, 'search'):
                exc_type, e, traceback = sys.exc_info()
                raise exc_type, e.message, traceback



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging
from operator import itemgetter, attrgetter
from collections import OrderedDict
from itertools import chain
try:
   from hashlib import sha1 as sha
except ImportError:
    from sha import sha

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.fields import FieldDoesNotExist
try:
    from django.db.models.sql.constants import LOOKUP_SEP
except:
    from django.db.models.constants import LOOKUP_SEP
from django.core.exceptions import ObjectDoesNotExist, ImproperlyConfigured
from django.utils.text import capfirst
from django.forms.forms import pretty_name

from .ormutils import resolve_orm_path


log = logging.getLogger(__name__)

# Base fields to detect built-in operator types.  Fields that subclass these are implicitly included
# in the type check.
TEXT_FIELDS = (models.CharField, models.TextField, models.FilePathField, models.IPAddressField,
               models.GenericIPAddressField, models.FileField)
DATE_FIELDS = (models.DateField, models.TimeField)
NUMERIC_FIELDS = (models.IntegerField, models.AutoField, models.FloatField, models.DecimalField)
BOOLEAN_FIELDS = (models.BooleanField, models.NullBooleanField)
RELATIONSHIP_FIELDS = (models.ForeignKey, models.ManyToManyField, models.OneToOneField)

# Maps the core field types to the default available search operators.
# The ORM query representation (e.g., "icontains") is not sent to the frontend.
# Notably, entries starting with  "!" are negative filters
OPERATOR_MAP = {
    'text': (
        ('icontains', "contains"),
        ('!icontains', "doesn't contain"),
        ('iexact', "= equal"),
        ('!iexact', "≠ not equal"),
        ('!isnull', "exists"),
        ('isnull', "doesn't exist"),
    ),
    'date': (
        ('exact', "= equal"),
        ('!exact', "≠ not equal"),
        ('gt', "> greater than"),
        ('lt', "< less than"),
        ('range', "between"),
        ('!isnull', "exists"),
        ('isnull', "doesn't exist"),
    ),
    'number': (
        ('exact', "= equal"),
        ('gt', "> greater than"),
        ('lt', "< less than"),
        ('range', "between"),
        ('!isnull', "exists"),
        ('isnull', "doesn't exist"),
        ('icontains', "contains"),
        ('!icontains', "doesn't contain"),
    ),
    'boolean': (
        ('exact', "= equal"),
        ('!isnull', "exists"),
        ('isnull', "doesn't exist"),
    ),
    'model': (
        ('!isnull', "exists"),
        ('isnull', "doesn't exist"),
    ),

    # If a field defines a "choices" list, we can't get very fancy with operator types
    'choices': (
        ('exact', "is"),
    ),
}


class ModelSearch(object):
    """ Contains search and display configuration for a single Model. """

    model = None
    verbose_name = None
    verbose_name_plural = None

    display_fields = None
    search_fields = None

    _display_fields = None
    _fields = None

    def __init__(self, model):
        self.model = model
        if not self.verbose_name_plural:
            # If the plural name is unset, but the single name is, pluralize the single name instead
            # of reverting back to the model's Meta verbose_plural_name (which might just be a
            # pluralization of its verbose_name, which is explicitly being overridden).
            if self.verbose_name:
                self.verbose_name_plural = self.verbose_name + "s"
            else:
                self.verbose_name_plural = capfirst(self.model._meta.verbose_name_plural)
        if not self.verbose_name:
            self.verbose_name = capfirst(self.model._meta.verbose_name)

        # Read the configured fields
        self._process_display_fields()
        self._process_searchable_fields()

        # Determine the ContentType in advance.
        self._content_type = ContentType.objects.get_for_model(self.model)

        # Process the display fields
        if not self.display_fields:
            self.display_fields = list(map(attrgetter('name'), self.model._meta.local_fields))

    def _process_display_fields(self):
        """
        Converts a potentially uneven collection of strings and tuples in ``display_fields`` to
        a uniform list of 3-tuples in the form ("Verbose name", "field_name", field_instance)

        """

        self._display_fields = []
        for field_info in self.display_fields:
            if isinstance(field_info, (tuple, list)):
                verbose_name, field_name = field_info
            else:
                field_name = field_info
                verbose_name = None

            field = resolve_orm_path(self.model, field_name)

            if verbose_name is None:
                verbose_name = field.verbose_name

            self._display_fields.append((capfirst(verbose_name), field_name, field))
        return self._display_fields

    def _process_searchable_fields(self):
        """
        Crunches the information in ``display_fields`` and the intricate ``search_fields`` to
        produce a single consistent index of all searchable fields.

        The resulting index is formatted as an OrderedDict where keys are "friendly" readable names
        and are mapped to ORM-style paths such as "subdivision__name"

        """

        self._fields = OrderedDict()

        # Get flattened sequence of 3-tuples: ([orm_path,...], verbose_name, Field)
        extended_info = self._get_field_info([], self.model, None, self.search_fields)

        # Store each element's [:2] as the field choices
        self._fields.update(map(itemgetter(slice(2)), extended_info))

        # Store each element's [::2] (that is, [0] and [2]) as a mapping to the field object
        self.field_types = dict(map(itemgetter(slice(0, None, 2)), extended_info))




    def _get_field_info(self, orm_path_bits, model, related_name, field_list,
                        include_model_in_verbose_name=True):
        """
        Recurses the fields listed on the model to provide a complete index of their ORM paths and
        friendly names.
        """

        if related_name:
            if isinstance(related_name, basestring):

                try:
                    # Post 1.8
                    field = model._meta.get_field(related_name)
                    direct = not field.auto_created or field.concrete
                    related_model = None if field.model._meta.concrete_model is model else field.model._meta.concrete_model
                except FieldDoesNotExist:
                    # Pre 1.8
                    field, related_model, direct, m2m = model._meta.get_field_by_name(related_name)

                if related_model is None:  # Field is local, follow relationship to find the model
                    if direct:
                        if hasattr(field, 'field'):
                            related_model = field.field.rel.to
                        else:
                            related_model = field.rel.to
                    else:

                        if hasattr(field, 'related_model'):
                            # Post 1.8
                            related_model = field.related_model
                        else:
                            # Pre 1.8
                            related_model = field.model
                else:  # Field is a virtual field from reverse relation
                    if hasattr(field, 'related_model'):
                        # Post 1.8
                        related_model = field.related_model
                    else:
                        # Pre 1.8
                        related_model = field.model
            else:
                # Syntax allowing a model class reference itself.
                # Trace the relationship backwards from the related class's Meta to get the field
                # name.
                related_model = related_name

                try:
                    # Post 1.8
                    fk_objects = [
                        f for f in related_model._meta.get_fields()
                        if (f.one_to_many or f.one_to_one)
                           and f.auto_created and not f.concrete
                    ]
                    m2m_objects = [
                        f for f in related_model._meta.get_fields(include_hidden=True)
                        if f.many_to_many and f.auto_created
                    ]

                except:
                    # Pre 1.8
                    raise
                    fk_objects = related_model._meta.get_all_related_objects()
                    m2m_objects = related_model._meta.get_all_related_many_to_many_objects()

                for related_object in chain(fk_objects, m2m_objects):
                    if related_object.field.model is model:
                        related_name = related_object.field.name
                        break
                else:
                    raise ValueError("Relationship field from %r to %r not found." % (
                        model.__name__, related_model.__name__))
        else:
            related_model = model


        base_orm_path = orm_path_bits
        sub_fields = []

        for field_name in field_list:
            if isinstance(field_name, dict):
                # Nested dict.  Recurse the function with the new related context
                extended_name, sub_field_list = field_name.items()[0]
                orm_path_bits = base_orm_path[:]
                if related_name:
                    orm_path_bits.append(related_name)
                sub_fields.extend(self._get_field_info(orm_path_bits, related_model, extended_name,
                        sub_field_list))
            else:
                # Raw field name or 2-tuple

                field = None
                # Check for a 2-tuple of ("Friendly name", "field_name")
                if isinstance(field_name, (tuple, list)):
                    verbose_name, field_name = field_name
                else:
                    # Derive a verbose name.  If the field comes from a model other than the base
                    # model, prepend the relate model's own verbose name.
                    try:
                        # Post 1.8
                        field = related_model._meta.get_field(field_name)
                    except:
                        # Pre 1.8
                        field, _, _, _ = related_model._meta.get_field_by_name(field_name)

                    if related_name:
                        verbose_name_bits = [related_model._meta.verbose_name, field.verbose_name]
                    else:
                        verbose_name_bits = [field.verbose_name]
                    verbose_name = ' '.join(map(pretty_name, verbose_name_bits))

                if not isinstance(field_name, (tuple, list)):
                    field_name = (field_name,)
                if field is None:
                    field = resolve_orm_path(related_model, field_name[0])

                try:
                    field = field.field
                except:
                    pass

                orm_path_bits = base_orm_path[:]
                if related_name:
                    orm_path_bits.append(related_name)

                orm_info = tuple('__'.join(orm_path_bits + [component])
                        for component in field_name)
                sub_fields.append([orm_info, verbose_name, field])
                # print orm_info, field.name, related_model, verbose_name

            # print sub_fields

        return sub_fields

    def get_operator_choices(self, field=None, hash=None, flat=False):
        """
        Returns the sequence of 2-tuples of ('querytype', "Friendly Operator Name") for the given
        ``field`` Field instance, or given the ``hash`` representing that field.

        If ``flat``, then only the text from each option is returned (read: not the "value"),
        forming a sequence of strings such as ("= equal", "< less than", ...).

        """

        # TODO: Fields with a "choices" setting might be an IntegerField, but provide text-based
        # values that the user will intuitively use as search values.  A "choices" setting might
        # need to assume string data regardless of the underlying type.

        if hash is not None:
            field = self.reverse_field_hash(hash)
        elif field is None:
            return []

        field_type = self.field_types[field]
        classification = self.get_field_classification(field_type)
        choices = OPERATOR_MAP[classification]

        # Remove the 'isnull' and 'isnotnull' operators if this field instance can't be null anyway
        if not field_type.null:
            choices = filter(lambda c: c[0] not in ('isnull', '!isnull'), choices)

        if flat:
            choices = map(itemgetter(1), choices)

        return choices

    def get_field_classification(self, field):
        """
        Use field (either a proper Django ``Field`` instance or a field definition tuple from the
        configuration) to return a type label for the field's data type.

        Returns one of "text", "date", "number", "boolean", or "model" (for relationship fields).

        """

        if isinstance(field, tuple):
            field = self.field_types[field]

        if field.choices:
            return 'choices'
        elif isinstance(field, TEXT_FIELDS):
            return 'text'
        elif isinstance(field, DATE_FIELDS):
            return 'date'
        elif isinstance(field, NUMERIC_FIELDS):
            return 'number'
        elif isinstance(field, BOOLEAN_FIELDS):
            return 'boolean'
        elif isinstance(field, RELATIONSHIP_FIELDS):
            return 'model'
        else:
            raise ValueError("Unhandled field type %s" % field.__class__.__name__)

    def get_searchable_field_choices(self, include_types=False):
        """
        Returns an iterable of 2-tuples suitable for use as a form's ``choices`` attribute.

        The ORM path is obscured for use as the <option> tag values.

        """

        choices = self._fields.items()
        if include_types:
            choices = map(lambda c: c + (self.get_field_classification(c[0]),), choices)

        # Perform a sha hash on the ORM path to get something unique and obscured for the frontend
        encode_value = lambda pair: (sha(','.join(pair[0])).hexdigest(),) + tuple(pair[1:])
        return map(encode_value, choices)

    def reverse_field_hash(self, hash):
        """ Returns the hash of field ORM paths derived from the initial configuration. """

        hashes = map(itemgetter(0), self.get_searchable_field_choices())
        try:
            return self._fields.keys()[hashes.index(hash)]
        except IndexError:
            # raise ValueError("Unknown field hash")
            # TODO Fix me!
            log.exception("Unknown field hash")
            return None

    def get_display_fields(self):
        """ Returns the list of labels for the display fields. """
        return list(map(itemgetter(0), self._display_fields))

    def user_has_perm(self, user):
        """
        Returns ``True`` or ``False`` to indicate definitive user permission, or else ``None`` to
        let the registry default permission decide.

        """

        return None

    def get_queryset(self, request, user):
        """
        Hook for subclasses to modify querysets.  By default, this method returns the default
        manager's ``.all()`` method.  The request and user are also available for special
        requirements.

        """
        return self.model.objects.all()

    def get_object_data(self, obj):
        """
        Returns a list of values retrieved from ``obj``, automatically fetched according to the
        configured ``display_fields``.  Values that raise ``ObjectDoesNotExist`` or cause
        ``AttributeError`` (such as a null ForeignKey in the middle of a chain lookup) are
        suppressed and replaced with ``None``.

        """

        data = []
        for _, field_name, _ in self._display_fields:
            try:
                value = reduce(getattr, [obj] + field_name.split(LOOKUP_SEP))
            except (ObjectDoesNotExist, AttributeError):
                value = None

            if callable(value):
                value = value()
            data.append(value)

        # Convert the first column's data into a link to the model instance
        if hasattr(obj, 'get_absolute_url'):
            data[0] = "<a href='{}'>{}</a>".format(obj.get_absolute_url(), data[0])

        return data

class SearchRegistry(object):
    """
    Holds the gathered configurations from apps that use an appsearch.py module.

    """

    _registry = None
    permission = 'change_{}'

    def __init__(self):
        self._registry = {}

    def __iter__(self):
        for k in self._registry:
            yield k

    def __getitem__(self, k):
        if not isinstance(k, basestring):
            k = '.'.join((k._meta.app_label, k.__name__.lower()))
        return self._registry[k]

    def __contains__(self, k):
        try:
            config = self.__getitem__(k)
            return True
        except KeyError:
            return False

    def register(self, model, configuration):
        id_string = '.'.join((model._meta.app_label, model.__name__)).lower()
        log.debug("Registering %r for appsearch configuration class %r", id_string, configuration)
        self._registry[id_string] = configuration(model)

    def filter_configurations_by_permission(self, user, permission_code):
        configurations = self._registry.values()
        if permission_code is None:
            permission_code = self.permission

        def check_permission(config):
            permission = permission_code.format(config.model.__name__.lower())
            permission = '{}.{}'.format(config.model._meta.app_label, permission)
            return user.has_perm(permission)

        if user:
            configurations = filter(check_permission, configurations)

        return configurations

    def sort_function(self, configurations):
        return sorted(configurations, key=attrgetter('verbose_name'))

    def set_sort_function(self, f):
        self.sort_function = f

    def sort_configurations(self, configurations):
        return self.sort_function(configurations)

    def get_configurations(self, user=None, permission=None):
        """
        Hook for returning the registry data in a particular order.  By default, the configurations
        are returned in alphabetical order according to their model names.

        """

        configurations = self.filter_configurations_by_permission(user, permission)

        return self.sort_configurations(configurations)

    def get_configuration(self, model, user=None, permission=None):
        try:
            configuration = self[model]
        except KeyError:
            log.warn("No registered configuration for model %r.", model)
            configuration = None
        else:
            if user:
                available_configurations = self.filter_configurations_by_permission(user, permission)
                if configuration not in available_configurations:
                    log.warn("Configuration for model %r available, but user %r doesn't have "
                            "permission (permission filter: %r).", model, user.username,
                            permission or self.permission)
                    configuration = None

        return configuration

search = SearchRegistry()



# -*- coding: utf-8 -*-
"""utils.py: Searcher"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging
import json
from operator import itemgetter
from collections import defaultdict

from django.db.models.query import Q
from django.forms.formsets import formset_factory
from django.template import RequestContext
from django.template.loader import render_to_string
try:
    from django.db.models.sql.constants import LOOKUP_SEP
except:
    from django.db.models.constants import LOOKUP_SEP
from django.utils.safestring import mark_safe

from .registry import search, SearchRegistry
from .forms import ModelSelectionForm, ConstraintForm, ConstraintFormset
from .ormutils import resolve_orm_path

log = logging.getLogger(__name__)


class Searcher(object):
    """ Template helper, wrapping all the necessary components to render an appsearch page. """

    # Methods and fields not meant to be accessed from the template should start with an underscore
    # to let the template variable name resolution block access.

    # Customizable classes
    model_selection_form_class = ModelSelectionForm
    constraint_form_class = ConstraintForm
    constraint_formset_class = ConstraintFormset

    # Instance data
    model_selection_form = None
    constraint_formset = None

    results = None

    # Processing callback hooks
    _display_fields_callback = None
    _build_queryset_callback = None
    _process_results_callback = None

    ## Fallback items normally provided by the view
    context_object_name = 'search'

    # Default templates
    form_template_name = "appsearch/default_form.html"
    search_form_template_name = "appsearch/search_form.html"
    results_list_template_name = "appsearch/results_list.html"

    def __init__(self, request, url=None, querydict=None, registry=search, permission=None, **kwargs):
        self.kwargs = kwargs
        self.request = request
        self.url = url or request.path

        self._forms_ready = False
        self._set_up_forms(querydict or request.GET, registry, permission)
        self.permission = permission
        self.registry = registry

        # Fallback items
        self.context_object_name = kwargs.get('context_object_name', self.context_object_name)
        self.form_template_name = kwargs.get('form_template_name', self.form_template_name)
        self.search_form_template_name = kwargs.get('search_form_template_name', self.search_form_template_name)
        self.results_list_template_name = kwargs.get('results_list_template_name', self.results_list_template_name)

        self._display_fields_callback = kwargs.get('display_fields_callback')
        self._build_queryset_callback = kwargs.get('build_queryset_callback')
        self._process_results_callback = kwargs.get('process_results_callback')

    # Rendering methods
    def __unicode__(self):
        return render_to_string(self.form_template_name, RequestContext(self.request, {
            self.context_object_name: self,
        }))

    def render_search_form(self):
        """ Renders only the template at ``search_form_template_name`` """
        return render_to_string(self.search_form_template_name, RequestContext(self.request, {
            self.context_object_name: self,
        }))

    def render_results_list(self):
        """ Renders only the template at ``results_list_template_name`` """
        return render_to_string(self.results_list_template_name, RequestContext(self.request, {
            self.context_object_name: self,
        }))

    def render_constraint_fields(self, model):
        """ Renders into JSON the model's fields available for search queries. """

        configuration = self.registry.get_configuration(model)
        choices = configuration.get_searchable_field_choices(include_types=True)
        return json.dumps({'choices': choices})

    def get_constraint_field_operators(self, model, field=None, hash=None):
        """ Returns into JSON the model's field's valid search operators. """

        if field is None and hash is None:
            raise ValueError("Need one of 'field' or 'hash'.")

        configuration = self.registry.get_configuration(model)

        # Fetch the choices without the query orm language.  Only UI labels are used at this stage.
        choices = configuration.get_operator_choices(field=field, hash=hash, flat=True)
        return choices

    def render_all_constraint_choices(self):
        """
        Returns a mapping of all models to their field choices, and all models to field hashes to
        operators.

        """

        field_data = defaultdict(list)
        operator_data = defaultdict(dict)

        configurations = self.model_selection_form.configurations
        model_values = map(itemgetter(0), self.model_selection_form.fields['model'].choices)[1:]

        for model_value, config in zip(model_values, configurations):
            model = config.model
            field_choices = config.get_searchable_field_choices(include_types=True)
            for orm_hash, field_label, field_type in field_choices:
                operator_choices = self.get_constraint_field_operators(model, hash=orm_hash)
                field_data[model_value].append([orm_hash, field_label, field_type])
                operator_data[model_value][orm_hash] = operator_choices

        return mark_safe(json.dumps({
            'fields': field_data,
            'operators': operator_data,
        }))

    # Configuration methods
    def get_model_selection_form_class(self):
        """ Returns ``self.model_selection_form_class`` """
        return self.model_selection_form_class

    def get_constraint_form_class(self):
        """ Returns ``self.constraint_form_class`` """
        return self.constraint_form_class

    def get_constraint_formset_class(self):
        """ Returns ``self.constraint_formset_class`` """
        return self.constraint_formset_class

    @property
    def ready(self):
        """ Indicates if a search has been executed by the constructed state. """

        if self._forms_ready:
            self.model_config = self.model_selection_form.get_selected_configuration()
            self.model = self.model_config.model
        return self._forms_ready

    def _set_up_forms(self, querydict, registry, permission=None):
        ModelSelectionFormClass = self.get_model_selection_form_class()
        ConstraintFormClass = self.get_constraint_form_class()
        ConstraintFormsetClass = self.get_constraint_formset_class()
        ConstraintFormsetClass = formset_factory(ConstraintFormClass, formset=ConstraintFormsetClass)

        self.model_selection_form = ModelSelectionFormClass(registry, self.request.user, permission,
                querydict)

        if self.model_selection_form.is_valid():
            model_configuration = self.model_selection_form.get_selected_configuration()
            self.constraint_formset = ConstraintFormsetClass(model_configuration, querydict)
            if self.constraint_formset.is_valid():
                self._forms_ready = True
        else:
            self.model_selection_form = ModelSelectionFormClass(registry, self.request.user,
                    permission)
            self.constraint_formset = ConstraintFormsetClass(configuration=None)

    def _perform_search(self):
        """
        Generates the query using the validated constraint formset.  Also compiles a natural
        language string (``self.natural_string``) in the format of
        "N [Model]s with [field] [operator] '[term]'".

        """

        natural_string = []

        # 2-tuples of an operator and Q instance, sent through reduce() after it's built
        query_list = []

        for i, constraint_form in enumerate(self.constraint_formset):
            type_operator = constraint_form.cleaned_data['type']
            field_list = constraint_form.cleaned_data['field']
            constraint_operator = constraint_form.cleaned_data['operator']
            term = constraint_form.cleaned_data['term']
            end_term = constraint_form.cleaned_data['end_term']

            verbose_name = self.model_config._fields[field_list]

            # Build this field's query.
            # Search fields bound together in a tuple are considered OR conditions for a single
            # virtual field name.
            query = None

            # Iterate multiple fields defined in a compound column
            for field in field_list:
                value = term
                target_field = resolve_orm_path(self.model, field)

                # Prep an inverted lookup
                negative = constraint_operator.startswith('!')
                if negative:
                    constraint_operator = constraint_operator[1:]

                if constraint_operator == "isnull":
                    value = not negative
                    negative = False

                # Bake the queryset language string
                field_query = LOOKUP_SEP.join((field, constraint_operator))

                q = Q(**{
                    field_query: value,
                })
                log.debug("Querying %s [%d]: %s%s=%r", self.model.__name__, i, field_query,
                          "!" if negative else "", value)

                # Negate if necessary
                if negative:
                    q = ~q

                if query is None:
                    query = q
                else:
                    query |= q

            query_list.append((type_operator, query))

            # Do some natural processing
            if isinstance(value, (tuple, list)):
                value = ' - '.join(map(unicode, value))
            else:
                value = str(value)
            bits = [verbose_name, constraint_form['operator'].value(), value]
            if i != 0: # Skip the leading "and" on the first constraint
                bits.insert(0, constraint_form['type'].value())
            natural_string.append(bits)

        # The first query's "type" should be ignored for the sake of the reduce line below.
        query = reduce(lambda q1, (op, q2): op(q1, q2), query_list[1:], query_list[0][1])

        queryset = self.build_queryset(self.model, query)
        data_rows = self.process_results(queryset)

        self.results = {
            'count': len(queryset),
            'list': data_rows,
            'fields': self._get_display_fields(self.model, self.model_config),
            'natural_string': "where " + ', '.join(map(' '.join, natural_string)),
        }

    def _get_display_fields(self, model, config):
        if self._display_fields_callback:
            return self._display_fields_callback(self, model, config)
        return config.get_display_fields()

    def get_select_related_fields(self, model, config):
        """ Returns a list of queryset language names to pass into ``.select_related()`` """
        display_fields = config._display_fields
        related_names = set()
        for _, field_name, _ in display_fields:
            name_bits = field_name.rsplit(LOOKUP_SEP, 1)
            if len(name_bits) == 2:
                related_names.add(name_bits[0])

        return related_names

    def build_queryset(self, model, query, queryset=None):
        """
        Returns the queryset using ``query``.

        Default behavior inspects the display fields for any related items and requests their
        selection and adds ``.distinct()``.

        If ``base_queryset`` is provided, it will be used as the starting point for applying the
        ``query`` filter.  This is useful for subclasses that want to change the default manager
        being accessed for the initial queryset.  Passing up a ``queryset`` via super() will
        accomplish this, returning a fully built queryset with minimum hassle.

        If ``build_queryset`` callback was defined on the originating view, it will be called
        after the queryset is built and will be sent the searcher instance, model, config, initial
        ``Q`` query, and the derived queryset.  The callback should return the queryset in its final
        state for execution.

        """

        related_names = self.get_select_related_fields(model, self.model_config)

        if queryset is None:
            queryset = self.model_config.get_queryset(self.request, self.request.user)

        queryset = queryset.filter(query).select_related(*related_names).distinct()

        if self._build_queryset_callback:
            queryset = self._build_queryset_callback(self, model, self.model_config, query,
                    queryset)

        return queryset

    def process_results(self, queryset):
        """
        Should return a sequence of n-tuples, where ``n`` is the number of columns to be shown in
        the table.

        """

        if self._process_results_callback:
            return self._process_results_callback(self, self.model, self.model_config, queryset)

        return [self.model_config.get_object_data(obj) for obj in queryset]



# -*- coding: utf-8 -*-
"""ormutils.py: ORM Utils"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

try:
    from django.db.models.sql.constants import LOOKUP_SEP
except:
    from django.db.models.constants import LOOKUP_SEP
from django.db.models.fields import FieldDoesNotExist

def resolve_orm_path(model, orm_path):
    """
    Follows the queryset-style query path of ``orm_path`` starting from ``model`` class.  If the
    path ends up referring to a bad field name, ``django.db.models.fields.FieldDoesNotExist`` will
    be raised.

    """

    bits = orm_path.split(LOOKUP_SEP)
    endpoint_model = reduce(get_model_at_related_field, [model] + bits[:-1])
    try:
        # Post 1.8
        field = endpoint_model._meta.get_field(bits[-1])
    except:
        # Pre 1.8
        field, _, _, _ = endpoint_model._meta.get_field_by_name(bits[-1])
    return field


def get_model_at_related_field(model, attr):
    """
    Looks up ``attr`` as a field of ``model`` and returns the related model class.  If ``attr`` is
    not a relationship field, ``ValueError`` is raised.

    """

    try:
        field, _, direct, m2m = model._meta.get_field_by_name(attr)
    except FieldDoesNotExist:
        raise

    if not direct:
        if hasattr(field, 'related_model'):  # Reverse relationship
            # -- Django >=1.8 mode
            return field.related_model
        elif hasattr(field, "model"):
            # -- Django <1.8 mode
            return field.model

    if hasattr(field, 'rel') and field.rel:   # Forward/m2m relationship
        return field.rel.to

    if hasattr(field, 'field'):    # Forward GenericForeignKey in Django 1.6+
        return field.field.rel.to

    raise ValueError("{0}.{1} ({2}) is not a relationship field.".format(model.__name__, attr, field.__class__.__name__))


